@@ -28,22 +28,24 @@ char *program_name;
 /* SYNOPSIS */
 void usage(void)
 {
-   printf(  "USAGE:\t%s [options] <dir1> [<dir2> ...]\n\n", program_name );
+   printf(  "USAGE:\t%s [options] [<dir1> ...]\n\n", program_name );
    printf(  "\t-a\t   provide compatibility to freedups by William Stearns.[=-up]\n"
    	    "\t-c\t   count file space savings per linked file.\n"
 	    "\t-d\t   requires the modification time stamps to be equal.\n" 
 	    "\t-f\t   requires the path-stripped file names to be equal.\n"
+	    "\t-u\t   requires group to be equal.\n" 
 	    "\t-h\t   shows this help. [other option are ignored]\n"
 	    "\t-m <bytes> only touch larger files. (deprecated: use -o \"-size +#\")\n" 
 	    "\t-n\t   do not really perform links [no action].\n" );
    printf(  "\t-o <opts>  pass an option string to the initially called find command.\n"
 	    "\t-p\t   requires file permissions to be equal.\n" 
 	    "\t-s\t   generate symlinks although some given paths are relative.\n"
-	    "\t-u\t   requires user & group to be equal.\n" 
+	    "\t-u\t   requires user to be equal.\n" 
 	    "\t-v\t   display shell commands to perform linking [verbose].\n"
 	    "\t<dir>\t   any directory to scan for duplicate files recursively.\n\n" );
    printf(  "\tOptions are toggle switches. Their final state applies.\n"
 	    "\tLater <dir> entries are linked to the earlier ones.\n"
+	    "\tIf no <dir> is given, filenames are exepcted from standard input.\n"
 	    "\tVersion " VERSION " by Andreas Neuper (c)2007.\n\n"
          );
    return;
@@ -53,6 +55,7 @@ void usage(void)
 	int showsave=0;	/* -c */
 	int sametime=0;	/* -d */
 	int samename=0;	/* -f */
+	int samegrp =0;	/* -g */
 	int minbytes=0;	/* -m */
 	int testmode=0;	/* -n */
 	int sameperm=0;	/* -p */
@@ -142,7 +145,7 @@ static void gethash(finfo*a)
     fgets(buffer,MAXPATHLENGTH,file2hash);
     a->hash = calloc(MATCHCNT+2,sizeof(char));
     if( feof(file2hash) )
-      {
+    {
 	/*  hash was unsuccessful;
 	 *  usually fails due to special characters in the file name
 	 *  (this was at least the reason for redirecting errors)
@@ -154,8 +157,8 @@ static void gethash(finfo*a)
 	 *  to link it to some temporary file,
 	 *  to evaluate it, and to continue then
 	 */
-      } else
-      {
+    } else
+    {
 	/*  if everything works fine
 	 *  there should appear 32 characters
 	 *  followed by two spaces and the file name
@@ -165,7 +168,7 @@ static void gethash(finfo*a)
         if(0) printf("MD5=%s\n", buffer);
 	/* store the hash value to avoid recalculation */
         strncpy(a->hash,buffer,MATCHCNT);
-      }
+    }
     pclose(file2hash);
     /* the string was already initialised with zero
      * since I am somewhat paranoid ...
@@ -183,10 +186,10 @@ int statcpy(finfo*a)
     printf("%s(%x:%d:%d) --> \n", a->name, (int)astat->st_dev, (int)astat->st_ino, (int)astat->st_blocks);
 #endif /* DEBUG */
     if(lstat(a->name, astat) != 0)
-      {
+    {
 	perror("lstat() failed while reading file statistics");
 	return(-1);
-      }
+    }
 #ifdef DEBUG
     printf("\t\t%s(%x:%d:%d)\n", a->name, (int)astat->st_dev, (int)astat->st_ino, (int)astat->st_blocks);
 #endif /* DEBUG */
@@ -214,40 +217,40 @@ void link_files(finfo*a, finfo*b)
 	b->name, (int)b->info.st_size, b->hash);
 #endif /* DEBUG */
     if(! testmode )
-      {
+    {
 	/* move file to tmp ; link it ; remove tmp
 	 */
 	strcpy(tmpfile,b->name);
 	strcat(tmpfile,".tmp");
 	rename(b->name,tmpfile);
-      }
+    }
     if( a->info.st_dev == b->info.st_dev )
-      {
+    {
 	/* either st_dev or st_ino are different,
 	 * was checked earlier prior to call
 	 */
 	if( showmode) printf("ln \"%s\" \"%s\"\n",a->name,b->name);
 	if(!testmode) link(a->name,b->name);
         statcpy(b);
-      } else
-      {
+    } else
+    {
 	assert(!nosymlinks);
 	if( showmode) printf("ln -s \"%s\" \"%s\"\n",a->name,b->name);
 	if(!testmode) symlink(a->name,b->name);
 	symlinked++;
         statcpy(b);
-      }
+    }
     if( showsave) 
-      {
+    {
 	printf("Linking \"%s\" saves %d bytes\n",
 		basename(b->name), 
 		(int)(b->info).st_size);
-      }
+    }
     linksaved+=(int)(b->info).st_size;
     if(!testmode)
-      {
+    {
 	unlink(tmpfile);
-      }
+    }
     totalinked++;
     return;
 }
@@ -271,47 +274,58 @@ int comparison_by_content(const void*a,const  void*b)
     afile=fopen(((finfo*)a)->name,"r");
     bfile=fopen(((finfo*)b)->name,"r");
     while(!(feof(afile)||feof(bfile)))
-      {
+    {
 	aread=fread(ablock,BLOCK,1,afile);
 	bread=fread(bblock,BLOCK,1,bfile);
 	assert(aread==bread);
 	assert((aread!=0)||feof(afile));
 	result=memcmp(ablock,bblock,aread);
 	if( result!=0 )
-	  {
+	{
 	    fclose(afile);
 	    fclose(bfile);
 	    return(result);
-          }
-      }
+        }
+    }
     /* decide which has higher prio by
-     * 1) earlier named in command line
-     * NOT: 2) shorter path length /+/ may result in alternating linkage
+     * 1) collect links where more hardlinks exist
+     * 2) link to the one earlier named in command line
+     * 3) arbitrary /# inactive: shorter path name #/
+     * NOT: shorter path length /+/ may result in alternating linkage
      * the one with lower priority will be deleted
+     */
+    if(((finfo*)a)->info.st_nlink > ((finfo*)b)->info.st_nlink )
+    {
+	/* We should elimate unlinked files and
+	 * keep more often linked files as collectors
 	 */
-    if(((finfo*)a)->prio<((finfo*)b)->prio)
-      {
 	link_files(((finfo*)a),((finfo*)b));
-      } else
-      {
-	if(((finfo*)a)->prio>((finfo*)b)->prio)
-	  {
+    } else
+    {
+	if(((finfo*)a)->info.st_nlink < ((finfo*)b)->info.st_nlink )
+	{
 	    link_files(((finfo*)b),((finfo*)a));
-	  } else
-	  {
+	} else
+	{
+	    if(((finfo*)a)->prio<((finfo*)b)->prio)
+	    {
+		link_files(((finfo*)a),((finfo*)b));
+	    } else
+	    {
+		if(((finfo*)a)->prio>((finfo*)b)->prio)
+		{
+		    link_files(((finfo*)b),((finfo*)a));
+		} else
+		{
 /**/
 /* it did not work well to give priority within one file system tree	*/
 /*	    if(strlen(((finfo*)a)->name)<strlen(((finfo*)b)->name))	*/
 /**/
-	    /* We should elimate unnlinked files and
-             * keep more often linked files as collectors
-             */
-	    if(((finfo*)a)->info.st_nlink > ((finfo*)b)->info.st_nlink )
-	        link_files(((finfo*)a),((finfo*)b));
-	    else
-	        link_files(((finfo*)b),((finfo*)a));
-	  }
-      }
+		    link_files(((finfo*)a),((finfo*)b));
+		}
+	    }
+	}
+    }
     fclose(afile);
     fclose(bfile);
     return(0);
@@ -332,7 +346,8 @@ static int comparison_by_size(const void*a, const void*b)
     /* avoid work when not desired
      * follow the minbytes directive
      */
-    if(minbytes>(((finfo*)a)->info.st_size))
+    if((minbytes > (((finfo*)a)->info.st_size))
+    || (minbytes > (((finfo*)b)->info.st_size)))
       {
         return(sizediff);
       }
@@ -341,92 +356,91 @@ static int comparison_by_size(const void*a, const void*b)
      * - same size -> differs file?
      */
     if(sizediff == 0)
-      {
+    {
 #ifdef DEBUG
     printf("COMPARE: %x:%d <> %x:%d\n",
 	(int)(((finfo*)a)->info.st_dev),(int)(((finfo*)a)->info.st_ino),
 	(int)(((finfo*)b)->info.st_dev),(int)(((finfo*)b)->info.st_ino));
 #endif /* DEBUG */
-	/* softlink must not result in endless link loops
-	 */
+	/* softlink must not result in endless link loops */
 	if( (((finfo*)a)->info.st_mode & S_IFMT) != S_IFREG 
 	 || (((finfo*)b)->info.st_mode & S_IFMT) != S_IFREG )
-	    /* same size but no linking allowed
-	     */
+	    /* same size but no linking allowed */
 	    return 0;
 	/* compare the path-less file names
-	 * if the samename directive requires this
-	 */
+	 * if the samename directive requires this */
 	if( strcmp(basename(((finfo*)a)->name),basename(((finfo*)b)->name)) 
 	 && (samename==1) )
-	    /* different names do not allow linking
-	     */
+	    /* different names do not allow linking */
 	    return 0;
-	/* check whether the user and group entry match
-	 */
+	/* check whether the user entries match */
 	if( ((((finfo*)a)->info.st_uid)==(((finfo*)b)->info.st_uid)) 
-	 && ((((finfo*)a)->info.st_gid)==(((finfo*)b)->info.st_gid))
 	 &&  (sameuser==1) )
-	    /* different names do not allow linking
-	     */
+	    /* different user ids do not allow linking */
 	    return 0;
-	/* check whether the permissions match
-	 */
+	/* check whether the group entries match */
+	if( ((((finfo*)a)->info.st_gid)==(((finfo*)b)->info.st_gid)) 
+	 &&  (samegrp ==1) )
+	    /* different group ids do not allow linking */
+	    return 0;
+	/* check whether the permissions match */
 	if( ((((finfo*)a)->info.st_mode&S_IRWXUGO)==(((finfo*)b)->info.st_mode&S_IRWXUGO)) 
 	 && (sameperm==1) )
-	    /* different names do not allow linking
-	     */
+	    /* different permissions do not allow linking */
 	    return 0;
-	/* check whether the modification dates match
-	 */
+	/* check whether the modification dates match */
 	if( ((((finfo*)a)->info.st_mtime)==(((finfo*)b)->info.st_mtime)) 
 	 && (sametime==1) )
-	    /* different names do not allow linking
-	     */
+	    /* different modification time do not allow linking */
 	    return 0;
-	/* check whether already hardlinked
-	 */
-	if( (((finfo*)a)->info.st_ino != ((finfo*)b)->info.st_ino)
-	 || (((finfo*)a)->info.st_dev != ((finfo*)b)->info.st_dev) )
-	  {
+	/* check whether already hardlinked */
+	if( ((((finfo*)a)->info.st_ino != ((finfo*)b)->info.st_ino))
+	 || ((((finfo*)a)->info.st_dev != ((finfo*)b)->info.st_dev)) )
+	{
 #ifdef MD5SUM
 	    if((((finfo*)a)->hash)==NULL) gethash((finfo*)a);
 	    if((((finfo*)b)->hash)==NULL) gethash((finfo*)b);
 	    assert(((finfo*)a)->hash != NULL);
 	    assert(((finfo*)b)->hash != NULL);
 	    if( strncmp( (((finfo*)a)->hash), (((finfo*)b)->hash), MATCHCNT )!=0 )
-	      {
+	    {
 		hashreject++;
-	      } else
+	    } else
 #endif
-	      {
+	    {
 		comparison_by_content(a,b);
-	      }
-          }
-      }
+	    }
+	}
+    }
     return(sizediff);
 }
 
-#ifdef DEBUG
 void print_infolist(void)
 {
-  int n;
-  for(n=0;n<filecount;n++)
+    int n;
+    for(n=0;n<filecount;n++)
     {
-      printf("%9d %x:%05d %s\n",
-	(int)file_info[n].info.st_size,
-	(int)file_info[n].info.st_dev,
-	(int)file_info[n].info.st_ino,
-	file_info[n].name);
+        printf("%9d %x:%05d %s\n",
+	  (int)file_info[n].info.st_size,
+	  (int)file_info[n].info.st_dev,
+	  (int)file_info[n].info.st_ino,
+	  file_info[n].name);
     }
 }
 
-cleanonexit()
+void printstats(void)
 {
-    if(findopta!=NULL)  free(findopta);
-    if(finfosize!=NULL) free(finfosize);
-}
+    fprintf(stderr,"%d file%s of %d %sreplaced by links.\nThe total size of replac%s files %s %lld bytes.\n",
+	totalinked,(totalinked==1)?"":"s",filecount,testmode?"will be ":"",
+	testmode?"able":"ed", testmode?"is":"was", linksaved);
+#ifdef MD5SUM
+    fprintf(stderr,"md5 hash algorithm had to read %d files to avoid %d file comparisons.\n",hashread,hashreject);
+#endif /* MD5SUM */
+#ifdef DEBUG
+    print_infolist();
 #endif /* DEBUG */
+}
+
 
 int readtree( finfo**fin, int*total, FILE*in)
 /*
@@ -455,9 +469,11 @@ int readtree( finfo**fin, int*total, FILE*in)
 		count--;
 		continue;
 	    }
-	    if( count > (*total)-2 )
+	    if( count+filecount > (*total)-2 )
 	    {
+#ifdef DEBUG
 		fprintf(stderr,"N(%d) > Filecount(%d) ==> expand sorting field\n",count,(*total));
+#endif
 		(*total) = (*total) + INITIALFINFOCOUNT;
 		(*fin) = realloc((*fin), sizeof(finfo) * ( (*total) +4 ) );
 	        assert((*fin)!=NULL);
@@ -470,7 +486,7 @@ int readtree( finfo**fin, int*total, FILE*in)
 int
 main (int argc, char *argv[])
 {
-  int c,i,lastlinked=0;
+  int c,i,lastlinked=0,remember=0;
   FILE*liste;
   char buffer[MAXPATHLENGTH];
   program_name = argv[0];
@@ -480,37 +496,39 @@ main (int argc, char *argv[])
 #ifdef DEBUG
   /* Locale is not used in current implementation
    * Time is not used in current implementation
-	 */
+   */
   start_time = time (NULL);
   printf ("cur_day_start = %s", ctime (&cur_day_start));
 #endif /* DEBUG */
   exit_status = 0;
 
-  while ((c = getopt(argc, argv, "acdfhm:no:psuv")) != EOF) {
-            switch (c) {
-                  case 'a': sameperm   = sameuser = 1; break;
-                  case 'c': showsave   = 1-showsave;   break;
-                  case 'd': sametime   = 1-sametime;   break;
-                  case 'f': samename   = 1-samename;   break;
-                  case 'h': usage();                 exit(0);
-                  case 'm': minbytes   = atoi(optarg); break;
-                  case 'n': testmode   = 1-testmode;   break;
-                  case 'o': findopta   = calloc( strlen(optarg)+2, sizeof(char) );
-			    strcpy( findopta, optarg );break;
-                  case 'p': sameperm   = 1-sameperm;   break;
-                  case 's': nosymlinks = 1-nosymlinks; break;
-                  case 'u': sameuser   = 1-sameuser;   break;
-                  case 'v': showmode   = 1-showmode;   break;
-                  default:
-			fprintf(stderr,"wrong option: \"%c\"\n",c);
-                        usage();
-                        exit(-1);
-                  }
-            }
+    while ((c = getopt(argc, argv, "acdfghm:no:psuv")) != EOF) 
+    {
+        switch (c)
+	{
+	  case 'a': sameperm   = sameuser = samegrp = 1; break;
+	  case 'c': showsave   = 1-showsave;   break;
+	  case 'd': sametime   = 1-sametime;   break;
+	  case 'f': samename   = 1-samename;   break;
+	  case 'g': samegrp    = 1-samegrp ;   break;
+	  case 'h': usage();                 exit(0);
+	  case 'm': minbytes   = atoi(optarg); break;
+	  case 'n': testmode   = 1-testmode;   break;
+	  case 'o': findopta   = calloc( strlen(optarg)+2, sizeof(char) );
+		    strcpy( findopta, optarg );break;
+	  case 'p': sameperm   = 1-sameperm;   break;
+	  case 's': nosymlinks = 1-nosymlinks; break;
+	  case 'u': sameuser   = 1-sameuser;   break;
+	  case 'v': showmode   = 1-showmode;   break;
+	  default:
+		fprintf(stderr,"wrong option: \"%c\"\n",c);
+		usage();
+		exit(-1);
+	}
+    }
 #ifdef DEBUG
   printf("ARGC:%d\nOPTIND:%d\n",argc,optind);
 #endif /* DEBUG */
-	 
   /* 
    * Collect files to scan.
    * Read them and fill in stat
@@ -518,15 +536,14 @@ main (int argc, char *argv[])
   file_info = (finfo*)calloc(finfosize+4,sizeof(finfo));
   assert(file_info!=NULL);
   if( argc == optind )
-    {
-      /*
-       * Read file names (not dir names) from stdin
-       */
-      fprintf(stderr,"Taking file names from stdin\n\n");
-      filecount = readtree( &file_info, &finfosize, stdin );
+    {	/*
+	 * Read file names (not dir names) from stdin
+	 */
+	fprintf(stderr,"Taking file names from stdin\n\n");
+	filecount = readtree( &file_info, &finfosize, stdin );
 	/*
-         * Now we will skip the following for-loop
- 	 */
+	 * Now we will skip the following for-loop
+	 */
     } else
     {
       for (i=optind,filecount=0;i < argc;i++)
@@ -542,23 +559,16 @@ main (int argc, char *argv[])
 	  pclose(liste);
 	}
     }
-  fprintf(stderr,"%d files to investigate\n\n",filecount);
-  do{
-      lastlinked=totalinked;
-      qsort( file_info, filecount, sizeof(finfo), comparison_by_size );
-#ifdef DEBUG
-      fprintf(stderr,"sorting loop ended with %d:%d\n",lastlinked,totalinked);
+    fprintf(stderr,"%d files to investigate\n\n",filecount);
+    do
+    {
+	remember=totalinked-lastlinked;
+        lastlinked=totalinked;
+        qsort( file_info, filecount, sizeof(finfo), comparison_by_size );
+#ifndef DEBUG
+        fprintf(stderr,"sorting loop ended with %d:%d\n",lastlinked,totalinked);
 #endif /* DEBUG */
-    }
-  while(lastlinked<totalinked && ! testmode);
-  fprintf(stderr,"%d file%s of %d %sreplaced by links.\nThe total size of replac%s files %s %lld bytes.\n",
-	totalinked,(totalinked==1)?"":"s",filecount,testmode?"will be ":"",
-	testmode?"able":"ed", testmode?"is":"was", linksaved);
-#ifdef MD5SUM
-  fprintf(stderr,"md5 hash algorithm had to read %d files to avoid %d file comparisons.\n",hashread,hashreject);
-#endif /* MD5SUM */
-#ifdef DEBUG
-  print_infolist();
-#endif /* DEBUG */
-  exit (exit_status);
+    } while(lastlinked<totalinked && remember!=(totalinked-lastlinked) && ! testmode);
+    printstats();
+    exit(exit_status);
 }
